#Le début du script gérant les polynômes
#Une programmation Python avancée encapsulerait ces fonctions dans une classe
#la programmation par classes, dite Orientée Objet, est hors programme BCPST
import numpy as np
import matplotlib.pyplot as plt


def degre(P):
    """
    retourne le degré du polynôme P
    Si P=0, retourne -1
    """
    #A compléter!!
    deg = len(P)-1
    while(deg >= 0 and P[deg] == 0) :#on parcourt la liste à partir de la fin
        deg = deg-1
    return deg

def trim(P):
    """
    retourne le codage de P le plus économique
    trim=tondre en anglais: on "tond" les éventuels zeros finaux
    """
    d = degre(P)
    if d == -1 :
        return [] #le polynome nul est codé par la liste vide
    else:
        return P[:d+1] #on utilise le slicing de tableaux

def coeff(P,k):
    """
    retourne p_k indépendamment du fait que k soit < len(P) ou >=
    """
    if k < len(P):
        return P[k]
    else:
        return 0 #si on est hors bornes, p_k=0

def somme(P, Q):
    """
    Retourne le polynome somme des deux polynomes
    """
    dp = degre(P); dq = degre(Q)
    somme = [] #la liste destinée à recevoir le résultat de la somme
    for k in range(max(dp,dq)+1):
        somme.append(coeff(P,k)+coeff(Q,k))
    return trim(somme)

def pdt_par_scalaire(lambada, P):
    """
    lambada: un nombre; P: un polynome. retourne lambada.P
    NB lambda est un mot réservé de Python, d'ou l'identifiant lambada
    """
    pdt = [] #la liste destinée à recevoir le résultat du produit
    for co in P : #Python permet de boucler sur les éléments d'une liste
        #A completer!!!!
        pdt.append(lambada*co)
    return pdt

def produitX(P):
    """
    retourne le polynome XP
    """
    return [0]+P #le produit revient simplement à décaler la liste 
                 #vers la droite en introduisant 0 comme premier terme
        
def produit(P, Q):
    """
    Retourne le polynome produit des deux polynomes P et Q
    """
    dp = degre(P); dq = degre(Q)
    pdt = [] #la liste destinée à recevoir le résultat du produit
    #cas du pol nul
    if(dp<0 or dq<0):
        return pdt
    for n in range(dp+dq+1):
        co = 0
        for k in range(n+1):
            co += coeff(P,k)*coeff(Q,n-k)
        pdt.append(co)
    return pdt

def derive(P):
    """
    retourne le polynome dérivé de P 
    """
    der = []
    dp = degre(P)
    if(dp<0): #le cas du polynome nul
        return []
    for k in range(dp):
        der.append((k+1)*P[k+1])
    return der

def subst_reel(P,x):
    """
    retourne le polynome P evalué en le nombre x
    x peut aussi être un ndarray, auquel cas 
      on applique P composante par composante
    """
    #on utilise la formule P(x)=((...)x*+p_1)*x+p_0
    valeur=0*x #on commence à zéro, permet d'avoir un ndarray si x en est un
    if P == []: #cas du polynome nul
        return valeur
    #on parcourt la liste à l'envers
    k = degre(P)
    while(k >=0):
        valeur = valeur*x + P[k]
        k = k-1
    return valeur


def subst_poly(P,Q):
    """
    retourne le polynome P evalué en le polynome Q, P(Q)=P \circ Q
    """
    #on utilise la formule P(Q)=((...)*Q+p_1)*Q+p_0
    comp=[] #on commence à zéro, permet d'avoir un ndarray si x en est un
    if P == []: #cas du polynome nul
        return comp
    #on parcourt la liste à l'envers
    k = degre(P)
    while(k >=0):
        comp = somme(produit(comp,Q),[P[k]])
        k = k-1
    return comp

def subst_matrice(P,A):
    """
    retourne le polynome P evalué en la matrice A, P(A)=p_0.I_n+p_1.A+...
    """
    #on utilise la formule P(Q)=((...)*Q+p_1)*Q+p_0
    In = np.eye(A.shape[0]) #la matrice identité de meme taille que A
    valeur = np.zeros(A) #on commence à zéro, 
                       #avec une matrice carrée de meme taille que A
    if P == []: #cas du polynome nul
        return valeur
    #on parcourt la liste à l'envers
    k = degre(P)
    while(k >=0):
        valeur = np.dot(valeur,A)+P[k]*In
        k = k-1
    return valeur

def Hermite(n=6):
    """
    retourne la liste des n premiers polynomes de Hermite H_0,..H_{n-1}
    applique la récurrence H du texte
    H_{n+1}=X.H_n - H'_n
    """
    H_liste = [[1]] #H_0 c'est le polynome constant 1, [1] dans notre notation
    for k in range(n):
        print("calcul de H_",k+1)
        H_liste.append(
            somme(
                produitX(H_liste[-1]),
                pdt_par_scalaire(-1, derive(H_liste[-1]))
            )
        )
    return H_liste

#Partie F
import numpy.linalg as la

#Q F.1
def Delta(n) :
    """
    Retourne la matrice Delta_n du texte
    """
    D = np.zeros((n,n))
    for k in range(n-1) :
        D[k, k+1] = np.sqrt(k+1)
        D[k+1, k] = np.sqrt(k+1) #D[k+1][k] est OK mais moins joli
    return D

#Q F.3
def EgaliteRacinesVp(n) :
    H_liste = Hermite(n)
    for k in range(1,n+1):
        D = Delta(k)
        alpha = la.eigvals(D)
        H_alpha = subst_reel(H_liste[k],alpha)
        print("k={} Racines ? {} Valeurs {} ".format(k, alpha, H_alpha))


if __name__=='__main__': #tests du module
    #le test #1 c'est calculer les 6 premiers polynomes de Hermite 
    #et de tracer leurs graphes sur [-5,5]
    H_liste = Hermite()
    plt.subplots()
    x = np.linspace(-5,5,100)
    for k in range(len(H_liste)):
        y = subst_reel(H_liste[k],x)
        plt.plot(x,y,label='$y=H_{'+str(k)+'}(x)$')
    plt.xlabel('x')
    plt.ylabel('y')
    plt.legend()
    plt.ylim((-10,10))
    plt.grid()
    plt.savefig('hermite6.pdf',format='pdf')
    plt.show()


    
